import io
from typing import Tuple, List

import asyncio
import httpx
import numpy as np
from PIL import Image, ImageEnhance
from nonebot import get_driver
from nonebot.log import logger

np.seterr(divide="ignore", invalid="ignore")
driver = get_driver()
client = httpx.AsyncClient()


def resize_image(
    im1: Image.Image, im2: Image.Image, mode: str
) -> Tuple[Image.Image, Image.Image]:
    """
    统一图像大小
    """
    im1_w, im1_h = im1.size
    if im1_w * im1_h > 1500 * 2000:
        if im1_w > 1500:
            im1 = im1.resize((1500, int(im1_h * (1500 / im1_w))))
        else:
            im1 = im1.resize((int(im1_w * (1500 / im1_h)), 1500))

    _wimg = im1.convert(mode)
    _bimg = im2.convert(mode).resize(_wimg.size, Image.Resampling.NEAREST)

    wwidth, wheight = _wimg.size
    bwidth, bheight = _bimg.size

    width = max(wwidth, bwidth)
    height = max(wheight, bheight)

    wimg = Image.new(mode, (width, height), 255)
    bimg = Image.new(mode, (width, height), 0)

    wimg.paste(_wimg, ((width - wwidth) // 2, (height - wheight) // 2))
    bimg.paste(_bimg, ((width - bwidth) // 2, (height - bheight) // 2))

    return wimg, bimg


def gray_car(
    wimg: Image.Image,
    bimg: Image.Image,
    chess: bool = False,
):
    """
    发黑白车
    :param wimg: 白色背景下的图片
    :param bimg: 黑色背景下的图片
    :param wlight: wimg 的亮度
    :param blight: bimg 的亮度
    :param chess: 是否棋盘格化
    :return: 处理后的图像
    """
    wimg, bimg = resize_image(wimg, bimg, "L")

    wpix = np.array(wimg).astype("float64")
    bpix = np.array(bimg).astype("float64")

    # 棋盘格化
    # 规则: if (x + y) % 2 == 0 { wpix[x][y] = 255 } else { bpix[x][y] = 0 }
    if chess:
        wpix[::2, ::2] = 255.0
        bpix[1::2, 1::2] = 0.0

    wpix = wpix * 0.5 + 128
    bpix *= 0.5

    a = 1.0 - wpix / 255.0 + bpix / 255.0
    r = np.where(abs(a) > 1e-6, bpix / a, 255.0)

    pixels = np.dstack((r, r, r, a * 255.0))

    pixels[pixels > 255] = 255

    output = io.BytesIO()
    Image.fromarray(pixels.astype("uint8"), "RGBA").save(output, format="png")
    return output


def color_car(
    wimg: Image.Image,
    bimg: Image.Image,
    wlight: float = 1.0,
    blight: float = 0.18,
    wcolor: float = 0.5,
    bcolor: float = 0.7,
    chess: bool = False,
):
    """
    发彩色车
    :param wimg: 白色背景下的图片
    :param bimg: 黑色背景下的图片
    :param wlight: wimg 的亮度
    :param blight: bimg 的亮度
    :param wcolor: wimg 的色彩保留比例
    :param bcolor: bimg 的色彩保留比例
    :param chess: 是否棋盘格化
    :return: 处理后的图像
    """
    wimg = ImageEnhance.Brightness(wimg).enhance(wlight)
    bimg = ImageEnhance.Brightness(bimg).enhance(blight)

    wimg, bimg = resize_image(wimg, bimg, "RGB")

    wpix = np.array(wimg).astype("float64")
    bpix = np.array(bimg).astype("float64")

    if chess:
        wpix[::2, ::2] = [255.0, 255.0, 255.0]
        bpix[1::2, 1::2] = [0.0, 0.0, 0.0]

    wpix /= 255.0
    bpix /= 255.0

    wgray = wpix[:, :, 0] * 0.334 + wpix[:, :, 1] * 0.333 + wpix[:, :, 2] * 0.333
    wpix *= wcolor
    wpix[:, :, 0] += wgray * (1.0 - wcolor)
    wpix[:, :, 1] += wgray * (1.0 - wcolor)
    wpix[:, :, 2] += wgray * (1.0 - wcolor)

    bgray = bpix[:, :, 0] * 0.334 + bpix[:, :, 1] * 0.333 + bpix[:, :, 2] * 0.333
    bpix *= bcolor
    bpix[:, :, 0] += bgray * (1.0 - bcolor)
    bpix[:, :, 1] += bgray * (1.0 - bcolor)
    bpix[:, :, 2] += bgray * (1.0 - bcolor)

    d = 1.0 - wpix + bpix

    d[:, :, 0] = d[:, :, 1] = d[:, :, 2] = (
        d[:, :, 0] * 0.222 + d[:, :, 1] * 0.707 + d[:, :, 2] * 0.071
    )

    p = np.where(abs(d) > 1e-6, bpix / d * 255.0, 255.0)
    a = d[:, :, 0] * 255.0

    colors = np.zeros((p.shape[0], p.shape[1], 4))
    colors[:, :, :3] = p
    colors[:, :, -1] = a

    colors[colors > 255] = 255

    output = io.BytesIO()
    Image.fromarray(colors.astype("uint8")).convert("RGBA").save(output, format="png")
    return output


async def _download_img(url: str):
    r = await client.get(url, timeout=15)
    if r.status_code == 200:
        return Image.open(io.BytesIO(r.content))
    logger.warning(f"下载图片 {url} 失败: {r.status_code}")


async def get_imgs(img_urls: List[str]) -> List[Image.Image | None]:
    if not img_urls:
        return []
    imgs = await asyncio.gather(*[_download_img(url) for url in img_urls])
    return [img for img in imgs]


def seperate(img: Image.Image, bright_factor: float = 3.3):
    """
    返回 表图，里图
    """

    black_bg = Image.new("RGBA", img.size, (0, 0, 0, 0))
    white_bg = Image.new("RGBA", img.size, (255, 255, 255, 0))
    black_bg.paste(img, mask=img)
    white_bg.paste(img, mask=img)
    black_bg = ImageEnhance.Brightness(black_bg).enhance(bright_factor)
    out_o = io.BytesIO()
    out_i = io.BytesIO()
    white_bg.convert("RGB").save(out_o, format="jpeg")
    black_bg.convert("RGB").save(out_i, format="jpeg")

    return out_o, out_i


@driver.on_shutdown
async def _():
    await client.aclose()



import re
from nonebot.adapters import Bot, Event
from nonebot.params import Depends

from nonebot_plugin_alconna.uniseg import reply_fetch
from nonebot_plugin_alconna import Image, UniMessage


def Images():
    """
    消息包含的图片
    
    支持获取回复的消息中的图片
    """
    async def d(bot: Bot, event: Event):
        reply = await reply_fetch(event, bot)
        msg = UniMessage.generate_without_reply(event=event, bot=bot)
        if reply:
            msg.extend(UniMessage.generate_without_reply(message=reply.msg))  # type: ignore
        return msg.get(Image)

    return Depends(d)


def Mode():
    """
    幻影坦克合成模式: gray / color

    支持从回复消息中获取
    """
    async def d(bot: Bot, event: Event):
        reply = await reply_fetch(event, bot)
        m = ""
        if reply and reply.msg:
            m = (
                reply.msg.strip()
                if isinstance(reply.msg, str)
                else reply.msg.extract_plain_text().strip()
            )
        return event.get_plaintext().strip() or m

    return Depends(d)


def Bright():
    async def d(bot: Bot, event: Event):
        reply = await reply_fetch(event, bot)
        b = ""
        if reply and reply.msg:
            b = (
                reply.msg.strip()
                if isinstance(reply.msg, str)
                else reply.msg.extract_plain_text().strip()
            )
        matched_nums = re.search(r"\d+.?\d+", event.get_plaintext() + " " + b)
        if matched_nums:
            bright = float(matched_nums.group())
            if bright > 6:
                bright = 6.0
            return bright
        return 2.0

    return Depends(d)


def get_img_urls(msg_imgs: list[Image] = []):
    img_urls: list[str] = []
    if msg_imgs:
        img_urls.extend(img.url for img in msg_imgs if img.url)
    return img_urls



from nonebot import require, on_command
from nonebot.adapters import Message
from nonebot.plugin import PluginMetadata, inherit_supported_adapters
from nonebot.log import logger
from nonebot.params import CommandArg

require("nonebot_plugin_alconna")
require("nonebot_plugin_waiter")

from nonebot_plugin_alconna import Image, UniMessage
from nonebot_plugin_waiter import waiter

from .data_source import color_car, get_imgs, gray_car, seperate
from .depends import Images, Mode, Bright, get_img_urls


__plugin_meta__ = PluginMetadata(
    name="miragetank",
    description="合成/分离幻影坦克图片",
    usage="""
/miragetank <图片1> <图片2>
/分离幻影坦克 <图片> [亮度增强值]

可选参数：
    亮度增强值：取值建议 1~6，默认2（对应本插件合成的gray模式图，color模式图建议设置为5.5）
    """.strip(),
    type="application",
    homepage="https://github.com/1umine/nonebot_plugin_miragetank",
    supported_adapters=inherit_supported_adapters("nonebot_plugin_alconna"),
)

PRIORITY = 27
BLOCK = True
mirage_tank = on_command(
    "生成幻影坦克",
    aliases={"miragetank", "幻影坦克", "合成幻影坦克"},
    priority=PRIORITY,
    block=BLOCK,
)
sep_miragetank = on_command("分离幻影坦克", priority=PRIORITY, block=BLOCK)


@mirage_tank.handle()
async def _(
    matched_imgs: list[Image] = Images(),
    mode: Message = CommandArg(),
):
    img_urls = get_img_urls(msg_imgs=matched_imgs)
    generate_mode = mode.extract_plain_text().strip()
    if not generate_mode or not generate_mode in ("gray", "color"):
        await mirage_tank.send("请输入合成模式: gray 或 color")
    elif len(img_urls) < 2:
        await mirage_tank.send(f"还需要 {2 - len(img_urls)} 张图")

    @waiter(waits=["message"], keep_session=True)
    async def get_params(m: str = Mode(), imgs: list[Image] = Images()):
        return m, imgs

    async for r in get_params(retry=5, prompt=""):  # type: ignore
        r: tuple[str, list[Image]]
        m, imgs = r
        if m and m.strip() in ("取消", "结束", "算了"):
            await mirage_tank.finish("已取消")

        if not generate_mode and m and m.strip() in ("gray", "color"):
            generate_mode = m
        elif not generate_mode:
            await mirage_tank.send("请输入合成模式: gray/color (二选一)")

        if imgs and len(img_urls) < 2:
            img_urls.extend(img.url for img in imgs if img.url)

        if len(img_urls) < 2:
            await mirage_tank.send(f"还需要 {2 - len(img_urls)} 张图")

        elif generate_mode:
            break

    await mirage_tank.send("开始合成...")
    wimg, bimg = await get_imgs(img_urls[:2])
    if not wimg:
        await mirage_tank.finish("表图下载失败")
    if not bimg:
        await mirage_tank.finish("里图下载失败")

    if generate_mode == "color":
        await UniMessage.image(raw=color_car(wimg, bimg)).send()
    elif generate_mode == "gray":
        await UniMessage.image(raw=gray_car(wimg, bimg)).send()


# 分离幻影坦克
@sep_miragetank.handle()
async def _(
    images: list[Image] = Images(),
    bright: float = Bright(),
):
    img_urls = get_img_urls(msg_imgs=images)
    if not img_urls:
        await sep_miragetank.send("请发送一张幻影坦克图片")

        @waiter(waits=["message"], keep_session=True)
        async def get_img(imgs: list[Image] = Images()):
            img_urls.extend(get_img_urls(msg_imgs=imgs))
            return img_urls

        async for r in get_img(retry=2, prompt="请发送一张幻影坦克图片"):
            if r:
                break
        else:
            if not img_urls:
                await sep_miragetank.finish("已终止")

    await sep_miragetank.send("稍等，正在分离")
    img = (await get_imgs(img_urls[:1]))[0]
    if not img:
        await sep_miragetank.finish("图片下载失败")

    if img.format != "PNG":
        await sep_miragetank.finish(f"图片格式为 {img.format or '未知'}, 需要 PNG")
    try:
        outer, inner = seperate(img, bright_factor=bright)
    except Exception as e:
        logger.error(f"分离幻影坦克失败：{e}")
        await UniMessage.text("分离失败，请稍后再试").send(at_sender=True)
    await UniMessage.image(raw=outer).image(raw=inner).send(at_sender=True)
